using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.IO;
using System.Xml;
using Microsoft.DirectX.DirectInput;

 

namespace TerrainViewer
{ 

	public class WinForm : System.Windows.Forms.Form
	{
		public string appName = "3D Terrain Viewer";
		public string appVersion = "1.0 beta";

		private Microsoft.DirectX.Direct3D.Device device;
		private Microsoft.DirectX.DirectInput.Device keyb;
		private System.ComponentModel.Container components = null;
		private System.Windows.Forms.MainMenu mainMenu;
		private System.Windows.Forms.MenuItem menuItemFile;
		private System.Windows.Forms.MenuItem menuItemQuit;
		private System.Windows.Forms.MenuItem menuItemMaps;

		private float angle =  (float)(Math.PI / 2);
		private float angle2 = (float)(Math.PI / 4);
		private float dx = 0;
		private float dy = 0;
		private float dist = 170;
		float fov = (float)Math.PI/4;
		private bool spin = false;
		private Mesh mesh;
		private Texture texture;
		private Material material;
		public string mapName = "";
		public string mapListFileName = "maps.xml";
		public string textureFileName;
		public string terrainFileName;
		public float verticalFactor = 0.1f;
		Color fogColor = Color.FromArgb(208, 208, 208);

		private Point mouseDownStartPosition = Point.Empty;
		private float mouseDownStartDx;
		private float mouseDownStartDy;
		private float mouseDownStartAngle;
		private float mouseDownStartAngle2;


		public WinForm()
		{
			InitializeComponent();
			this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque, true);
		}
 

		public void InitializeDevice()
		{
			PresentParameters presentParams = new PresentParameters();
			presentParams.Windowed = true;
			presentParams.SwapEffect = SwapEffect.Discard;
			presentParams.AutoDepthStencilFormat = DepthFormat.D16;
			presentParams.EnableAutoDepthStencil = true;
			presentParams.PresentationInterval = PresentInterval.Immediate;
			device = new Microsoft.DirectX.Direct3D.Device(0, Microsoft.DirectX.Direct3D.DeviceType.Hardware, this, CreateFlags.HardwareVertexProcessing, presentParams);
			device.RenderState.FillMode = FillMode.Solid;
			device.DeviceResizing += new CancelEventHandler(OnDeviceResizing);
		}

		private void OnDeviceResizing(object sender, CancelEventArgs e)
		{
		}

		public void InitializeMapList()
		{
			XmlTextReader reader = null;
			// Load the reader with the data file and ignore all white space nodes.         
			reader = new XmlTextReader(mapListFileName);
			reader.WhitespaceHandling = WhitespaceHandling.None;

			// Parse the file.
			while (reader.Read()) 
			{
				switch (reader.NodeType) 
				{
					case XmlNodeType.Element:
					switch(reader.Name)
					{
						case "map" :
							if(reader["name"] != null)	// Add to Maps menu
							{
								MenuItem mi = new MenuItem(reader["name"]);
								mi.Click += new System.EventHandler(this.menuItemMaps_Click);
								menuItemMaps.MenuItems.Add(mi);			
							}
							break;
					}
						break;
				}       
			}  
			reader.Close();
		}

		private void menuItemMaps_Click(object sender, System.EventArgs e)
		{
			// Switch maps here
			MenuItem mapItem = (MenuItem)sender;
			string name = mapItem.Text;
			XmlTextReader reader = null;
			// Load the reader with the data file and ignore all white space nodes.         
			reader = new XmlTextReader(mapListFileName);
			reader.WhitespaceHandling = WhitespaceHandling.None;

			bool mapFound = false;
			// Parse the file.
			while (reader.Read()) 
			{
				switch (reader.NodeType) 
				{
					case XmlNodeType.Element:
					switch(reader.Name)
					{
						case "map" :
							if(reader["name"] != null)
							{
								if(reader["name"] == name) // Map found: switching
								{
									mapName = name;
									mapFound = true;
								}
								else
								{
									mapFound = false;
								}
							
							}
							break;
						case "terrain" :
							if(mapFound) 
							{
								if(reader["demfile"] != null) terrainFileName = reader["demfile"];
								if(reader["texturefile"] != null) textureFileName = reader["texturefile"];
								if(reader["altscale"] != null) verticalFactor = float.Parse(reader["altscale"]) / 10;
								
							}
							break;
					}
						break;
				}       
			}  
			reader.Close();
			MapMenuClearCheck();
			mapItem.Checked = true;
			MeshDispose();
			MeshDeclaration();
			this.Text = mapName + " - " + appName;
		}

		// Clear all check marks in Map menu
		public void MapMenuClearCheck()
		{
			foreach(MenuItem mi in menuItemMaps.MenuItems) mi.Checked = false;
		}

		// Load a map from the Maps menu by its rank
		public void MapMenuSelectMap(int num)
		{
			foreach(MenuItem mi in menuItemMaps.MenuItems) 
			{
				if(mi.Index == num) { menuItemMaps_Click(mi, null); break; }
			}
		}

		// Load a map from the Maps menu by its name
		public void MapMenuSelectMap(string name)
		{
			foreach(MenuItem mi in menuItemMaps.MenuItems) 
			{
				if(mi.Text == name) { menuItemMaps_Click(mi, null); break; }
			}
		}

		public void InitializeKeyboard()
		{
			keyb = new Microsoft.DirectX.DirectInput.Device(SystemGuid.Keyboard);
			keyb.SetCooperativeLevel(this, CooperativeLevelFlags.Background | CooperativeLevelFlags.NonExclusive);
			keyb.Acquire();
		}

 

		private void CameraPositioning()
		{
			//device.Transform.Projection = Matrix.PerspectiveFovLH(fov, this.Width/this.Height, 1f, 1000f);
			device.Transform.Projection = Matrix.PerspectiveFovLH(fov, device.Viewport.Width/device.Viewport.Height, 1f, 1000f);
			device.Transform.View = Matrix.LookAtLH(new Vector3(0,0,dist), new Vector3(0,0,0), new Vector3(1,0,0));
			device.RenderState.Lighting = true;
			device.RenderState.CullMode = Cull.CounterClockwise;
		}


  
		private void MeshDeclaration()
		{
			mesh = TerrainTile(device, terrainFileName);
			texture = TextureLoader.FromFile(device, textureFileName);
			material = new Material();
			material.Diffuse = Color.FromArgb(0xff, 0xff, 0xff, 0xff);
			material.Ambient = Color.White;
			material.Specular = Color.White;
			material.SpecularSharpness = 30.0f;  // thin highlight (smaller = larger)			
		}

		private void MeshDispose()
		{
			if(mesh != null) 
			{
				mesh.Dispose();
				mesh = null;
			}
			if(texture != null) 
			{
				texture.Dispose();
				texture = null;
			}
		}

		protected void SetupLights()
		{
			// Use a white, directional light coming from over our right shoulder
			device.Lights[0].Diffuse = Color.White;
			device.Lights[0].Specular = Color.FromArgb(0x80, 0x80, 0x80); // soft highlight
			device.Lights[0].Type = LightType.Directional;
			device.Lights[0].Direction = new Vector3(-1, 0, -1);
			device.Lights[0].Update(); 
			device.Lights[0].Enabled = true;
			device.RenderState.SpecularEnable = false;

			// Add a little ambient light to the scene
			device.RenderState.Ambient = Color.FromArgb(0x60, 0x60, 0x60);
		}

		// Frame loop
		protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
		{
			device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, fogColor , 1.0f, 0);
			device.BeginScene();
			
			// Rendering here
			CameraPositioning();
			device.VertexFormat = CustomVertex.PositionNormalTextured.Format;
			device.RenderState.ZBufferEnable = true;

			// Orientation angle / angle2
			device.Transform.World = Matrix.Translation(dx,dy, 0);
			device.Transform.World *= Matrix.RotationZ(angle);
			device.Transform.World *= Matrix.RotationY(angle2);
			SetupLights();

			if(texture != null) device.SetTexture(0,texture);
 
			// Texture stage states.
			device.TextureState[0].ColorOperation = TextureOperation.Modulate ;
			device.TextureState[0].ColorArgument1 = TextureArgument.TextureColor;
			device.TextureState[0].ColorArgument2 = TextureArgument.Diffuse;
			// Texture filtering
			//device.SamplerState[0].MipFilter = TextureFilter.Linear;
			device.SamplerState[0].MinFilter = TextureFilter.Linear;
			device.SamplerState[0].MagFilter = TextureFilter.Linear;

			device.Material = material;

			// Fog
			device.RenderState.FogEnable = true;
			device.RenderState.FogColor = fogColor;
			device.RenderState.FogTableMode = FogMode.Linear;
			device.RenderState.FogStart = 60;
			device.RenderState.FogEnd = 600;

			if(mesh != null) mesh.DrawSubset(0);

			// Done rendering
			device.EndScene();
			device.Present();
			this.Invalidate();

			ReadKeyboard();

			if(spin) angle-=0.002f; // slow spin

		}

 
		private void ReadKeyboard()
		{
			KeyboardState keys = keyb.GetCurrentKeyboardState();
			bool shift = keys[Key.LeftShift]  || keys[Key.RightShift];
			bool ctrl = keys[Key.LeftControl]  || keys[Key.RightControl];
			double moveFactor = dist * 0.01f;
			// Toggle map spin
			if (keys[Key.Space])	
			{
				spin = !spin;
			}
			// Rotate map
			if (keys[Key.RightArrow] && !shift)	
			{
				angle -= 0.02f;
			}
			if (keys[Key.LeftArrow] && !shift)
			{
				angle += 0.02f;
			}
			if (keys[Key.UpArrow] && shift)
			{
				angle2 += 0.02f;
			}
			if (keys[Key.DownArrow] && shift)
			{
				angle2 -= 0.02f;
			}
			// Move map
			if (keys[Key.RightArrow] && shift)	
			{
				dx -= (float)(Math.Sin(angle) * moveFactor);
				dy -= (float)(Math.Cos(angle) * moveFactor);
			}
			if (keys[Key.LeftArrow] && shift)
			{
				dx += (float)(Math.Sin(angle) * moveFactor);
				dy += (float)(Math.Cos(angle) * moveFactor);
			}
			if (keys[Key.UpArrow] && !shift)
			{
				dx -= (float)(Math.Cos(angle) * moveFactor);
				dy += (float)(Math.Sin(angle) * moveFactor);
			}
			if (keys[Key.DownArrow] && !shift)
			{
				dx += (float)(Math.Cos(angle) * moveFactor);
				dy -= (float)(Math.Sin(angle) * moveFactor);
			}
			// Change Distance
			if (keys[Key.NumPadPlus] && !shift)	
			{
				dist -= dist * 0.02f;
			}
			if (keys[Key.NumPadMinus] && !shift)
			{
				dist += dist * 0.02f;
			}
			// Change FOV
			if (keys[Key.NumPadPlus] && shift)	
			{
				fov -= 0.01f;
			}
			if (keys[Key.NumPadMinus] && shift)
			{
				fov += 0.01f;
			}
		}

		/// <summary>
		/// Occurs when the mouse wheel moves while the control has focus.
		/// </summary>
		protected override void OnMouseWheel(MouseEventArgs e)
		{
			try
			{
				this.dist -= this.dist * ( e.Delta/3000.0f );
			}
			finally
			{
				// Call the base class's OnMouseWheel method so that registered delegates receive the event.
				base.OnMouseWheel(e);
			}
		}
		protected override void OnMouseDown(MouseEventArgs e)
		{
			mouseDownStartPosition.X = e.X;
			mouseDownStartPosition.Y = e.Y;
			mouseDownStartDx = dx;
			mouseDownStartDy = dy;
			mouseDownStartAngle = angle;
			mouseDownStartAngle2 = angle2;
		}

		protected override void OnMouseUp(MouseEventArgs e)
		{
			mouseDownStartPosition = Point.Empty;
		}

		protected override void OnMouseMove(MouseEventArgs e)
		{
			if(mouseDownStartPosition == Point.Empty)
				return;
			bool isMouseLeftButtonDown = ((int)e.Button & (int)MouseButtons.Left) != 0;
			bool isMouseRightButtonDown = ((int)e.Button & (int)MouseButtons.Right) != 0;
			int dxMouse = this.mouseDownStartPosition.X - e.X;
			int dyMouse = this.mouseDownStartPosition.Y - e.Y;
			if(isMouseLeftButtonDown) 
			{
				double moveFactor = dist * 0.001f;
				dx = mouseDownStartDx - (float)(Math.Sin(angle) * dxMouse * moveFactor) + (float)(Math.Cos(angle) * dyMouse * moveFactor);
				dy = mouseDownStartDy - (float)(Math.Cos(angle) * dxMouse * moveFactor) - (float)(Math.Sin(angle) * dyMouse * moveFactor);
			}
			if(isMouseRightButtonDown) 
			{
				double spinFactor = 0.003f;
				angle = mouseDownStartAngle + (float)(dxMouse * spinFactor);
				angle2 = mouseDownStartAngle2 + (float)(dyMouse * spinFactor);
			}
		}

		/// <summary>
		/// Creates a terrain elevation mesh / tile
		/// </summary>
		/// <param name="device">The current direct3D drawing device.</param>
		/// <param name="terrainFilePath">The elevation DEM bitmap</param>
		/// <returns></returns>
		/// <remarks>
		/// Number of vertices in the mesh will be width x height of elevation map<br/>
		/// Number of faces	: (width-1) * (height-1) * 2
		/// Number of Indexes	: Number of faces * 3;
		/// </remarks>
		private Mesh TerrainTile(Microsoft.DirectX.Direct3D.Device device, string terrainFilePath)
		{
			int squareSize = 1;

			// Open elevation bitmap
			Bitmap DEM;
			DEM = new Bitmap(terrainFilePath);
			int width = DEM.Width;
			int height = DEM.Height;
			int halfW = width / 2;
			int halfH = height / 2;
			int numVertices = halfW * halfH;
			int numFaces	= (halfW-1) * (halfH-1) * 2;
			int indexCount	= numFaces * 3;

			// Create mesh
			Mesh mesh = new Mesh(numFaces,numVertices,MeshFlags.Managed,CustomVertex.PositionNormalTextured.Format,device);

			// Get the original mesh vertex buffer.
			int [] ranks = new int[1];
			ranks[0] = mesh.NumberVertices;
			System.Array arr = mesh.VertexBuffer.Lock(0,typeof(CustomVertex.PositionNormalTextured),LockFlags.None,ranks);

			// Set the vertex buffer
			int vertIndex=0;
			int offsetX = (width - 1) * squareSize / 2;
			int offsetY = (height - 1) * squareSize / 2;
			for(int y = 0; y < height-1; y+=2)
			{
				double vy = y * squareSize - offsetY;
				for(int x = 0; x < width-1; x+=2)
				{
					CustomVertex.PositionNormalTextured pnt = new CustomVertex.PositionNormalTextured();
					double vx = x * squareSize - offsetX;
					Color p = DEM.GetPixel(x, y);
					float alt = (float)p.R;
					pnt.X = (float)vx;
					pnt.Y = (float)vy;
					pnt.Z = alt * verticalFactor;
					pnt.Tu = (float)((float)x / (float)(width -1));
					pnt.Tv = (float)((float)y / (float)(height -1));

					arr.SetValue(pnt,vertIndex++);
				}
			}

			mesh.VertexBuffer.Unlock();
			ranks[0]=indexCount;
			arr = mesh.LockIndexBuffer(typeof(short),LockFlags.None,ranks);
			int i=0;
			short bottomVertex = 0;
			short topVertex = 0;
			for(short y = 0; y < (height - 3); y+=2)
			{
				topVertex = (short)(halfW*y/2);
				bottomVertex = (short)(topVertex + halfW);
				for(int x = 0; x < (width - 3); x+=2)
				{
					arr.SetValue(bottomVertex,i++);
					arr.SetValue(topVertex,i++);
					arr.SetValue((short)(topVertex+1),i++);
					arr.SetValue(bottomVertex,i++);
					arr.SetValue((short)(topVertex+1),i++);
					arr.SetValue((short)(bottomVertex+1),i++);
					bottomVertex++;
					topVertex++;
				}
			}
			mesh.IndexBuffer.SetData(arr,0,LockFlags.None);
			mesh.ComputeNormals();

			return mesh;
		}
		/// <summary>
		/// Creates a PositionNormalTextured sphere centered on zero
		/// </summary>
		/// <param name="device">The current direct3D drawing device.</param>
		/// <param name="radius">The sphere's radius</param>
		/// <param name="slices">Number of slices (Horizontal resolution).</param>
		/// <param name="stacks">Number of stacks. (Vertical resolution)</param>
		/// <returns></returns>
		/// <remarks>
		/// Number of vertices in the sphere will be (slices+1)*(stacks+1)<br/>
		/// Number of faces	:slices*stacks*2
		/// Number of Indexes	: Number of faces * 3;
		/// </remarks>
		private Mesh TexturedSphere(Microsoft.DirectX.Direct3D.Device device, float radius, int slices, int stacks)
		{
			int numVertices = (slices+1)*(stacks+1);
			int numFaces	= slices*stacks*2;
			int indexCount	= numFaces * 3;

			Mesh mesh = new Mesh(numFaces,numVertices,MeshFlags.Managed,CustomVertex.PositionNormalTextured.Format,device);

			// Get the original sphere's vertex buffer.
			int [] ranks = new int[1];
			ranks[0] = mesh.NumberVertices;
			System.Array arr = mesh.VertexBuffer.Lock(0,typeof(CustomVertex.PositionNormalTextured),LockFlags.None,ranks);

			// Set the vertex buffer
			int vertIndex=0;
			CustomVertex.PositionNormalTextured pnt;
			Vector3 v;
			// south pole vertex
			pnt = new CustomVertex.PositionNormalTextured();
			v = SphericalToCartesian( -90, 180, radius);
			pnt.X = v.X;
			pnt.Y = v.Y;
			pnt.Z = v.Z;
			pnt.Tu = 0.5f;
			pnt.Tv = 1.0f;
			v.Normalize();
			pnt.Normal = v;
			arr.SetValue(pnt,vertIndex++);
			// stacks and slices
			for(int stack = 1; stack < stacks; stack++)
			{
				double latitude = -90 + ((float)stack/stacks*(float)180.0);
				for(int slice = 0; slice <= slices; slice++)
				{
					pnt = new CustomVertex.PositionNormalTextured();
					double longitude = 180 - ((float)slice/slices*(float)360);
					v = SphericalToCartesian( latitude, longitude, radius);
					pnt.X = v.X;
					pnt.Y = v.Y;
					pnt.Z = v.Z;
					pnt.Tu = (float)slice/slices;
					pnt.Tv = 1.0f-(float)stack/stacks;
					v.Normalize();
					pnt.Normal = v;
					arr.SetValue(pnt,vertIndex++);
				}
			}
			// north pole vertex
			pnt = new CustomVertex.PositionNormalTextured();
			v = SphericalToCartesian( 90, 180, radius);
			pnt.X = v.X;
			pnt.Y = v.Y;
			pnt.Z = v.Z;
			pnt.Tu = 0.5f;
			pnt.Tv = 0.0f;
			v.Normalize();
			pnt.Normal = v;
			arr.SetValue(pnt,vertIndex++);

			mesh.VertexBuffer.Unlock();
			ranks[0]=indexCount;
			arr = mesh.LockIndexBuffer(typeof(short),LockFlags.None,ranks);
			int i=0;
			// wrap south pole
			short bottomVertex = 0;
			short topVertex = 1;
			for(int y = 0; y < slices; y++)
			{
				arr.SetValue(bottomVertex,i++);
				arr.SetValue(topVertex,i++);
				arr.SetValue((short)(topVertex+1),i++);
				topVertex++;
			}
			
			// stacks and slices
			for(short x = 1; x < stacks-1; x++)
			{
				bottomVertex = (short)((slices+1) * (x-1) + 1);
				topVertex = (short)(bottomVertex + slices + 1);
				for(int y = 0; y < slices; y++)
				{
					arr.SetValue(bottomVertex,i++);
					arr.SetValue(topVertex,i++);
					arr.SetValue((short)(topVertex+1),i++);
					arr.SetValue(bottomVertex,i++);
					arr.SetValue((short)(topVertex+1),i++);
					arr.SetValue((short)(bottomVertex+1),i++);
					bottomVertex++;
					topVertex++;
				}
			}
			// wrap north pole
			bottomVertex = (short)((slices+1) * (stacks-2) + 1);
			topVertex = (short)(bottomVertex + slices + 1);
			for(int y = 0; y < slices; y++)
			{
				arr.SetValue(bottomVertex,i++);
				arr.SetValue(topVertex,i++);
				arr.SetValue((short)(bottomVertex+1),i++);
				bottomVertex++;
			}


			mesh.IndexBuffer.SetData(arr,0,LockFlags.None);
			//mesh.ComputeNormals();

			return mesh;
		}

		/// <summary>
		/// Converts position in spherical coordinates (lat/lon/altitude) to cartesian (XYZ) coordinates.
		/// </summary>
		/// <param name="latitude">Latitude in decimal degrees</param>
		/// <param name="longitude">Longitude in decimal degrees</param>
		/// <param name="radius">Radius (OBS: not altitude)</param>
		/// <returns>Coordinates converted to cartesian (XYZ)</returns>
		public static Vector3 SphericalToCartesian(
			double latitude,
			double longitude,
			double radius
			)
		{
			latitude *= System.Math.PI / 180.0f;
			longitude *= System.Math.PI /180.0f;

			double radCosLat = radius * Math.Cos(latitude);

			return new Vector3(
				(float)(radCosLat * Math.Cos(longitude)),
				(float)(radCosLat * Math.Sin(longitude)),
				(float)(radius * Math.Sin(latitude)) );
		}

		protected override void Dispose (bool disposing)
		{
			MeshDispose();
			if (disposing)
			{
				if (components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose(disposing);
		}
 
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.Size = new System.Drawing.Size(800,800);
			this.Text = mapName + " - " + appName;
			this.mainMenu = new System.Windows.Forms.MainMenu();
			this.menuItemFile = new System.Windows.Forms.MenuItem();
			this.menuItemMaps = new System.Windows.Forms.MenuItem();
			this.menuItemQuit = new System.Windows.Forms.MenuItem();
			// 
			// mainMenu
			// 
			this.mainMenu.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
				this.menuItemFile,
				this.menuItemMaps});
			// 
			// menuItemFile
			// 
			this.menuItemFile.Index = 0;
			this.menuItemFile.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
				this.menuItemQuit});
			this.menuItemFile.Text = "&File";
			// 
			// menuItemQuit
			// 
			this.menuItemQuit.Index = 0;
			this.menuItemQuit.Text = "E&xit\tAlt+F4";
			this.menuItemQuit.Click += new System.EventHandler(this.menuItemQuit_Click);

			// 
			// menuItemMaps
			// 
			this.menuItemMaps.Index = 1;
			//this.menuItemMaps.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {});
			this.menuItemMaps.Text = "&Maps";

			this.Menu = this.mainMenu;

		}
    
		private void menuItemQuit_Click(object sender, System.EventArgs e)
		{
			Close();
		}

		static void Main()
		{
			using (WinForm form = new WinForm())
			{        
				form.InitializeMapList();
				form.InitializeDevice();
				form.CameraPositioning();
				form.MapMenuSelectMap(0); // load first map
				form.InitializeKeyboard();
				form.Show();
				Application.Run(form);
			}
		}
	}
}
 